"""
RT130 object to hold and process RefTek data
"""
from datetime import datetime
import os
from pathlib import Path
from typing import Tuple, List, Union

import numpy as np
from obspy.core import Stream

from sohstationviewer.model.gps_point import GPSPoint
from sohstationviewer.model.reftek.from_rt2ms import (
    core, soh_packet, packet)
from sohstationviewer.model.reftek.log_info import LogInfo
from sohstationviewer.model.data_type_model import DataTypeModel, ThreadStopped
from sohstationviewer.model.handling_data import (
    read_waveform_reftek, squash_gaps, sort_data, read_mp_trace, read_text)
from sohstationviewer.conf import constants
from sohstationviewer.controller.util import validate_file
from sohstationviewer.view.util.enums import LogType


class RT130(DataTypeModel):
    """
    read and process reftek file into object with properties can be used to
    plot SOH data, mass position data, waveform data and gaps
    """
    def __init__(self, *args, **kwarg):
        self.EH = {}
        super().__init__(*args, **kwarg)
        self.keys = set()
        self.req_data_streams: List[Union[int, str]] = self.req_wf_chans
        self.mass_pos_stream = {}
        self.found_data_streams = []

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.read_soh_index_waveform(self.dir)

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        self.selected_key = self.select_key()
        if self.selected_key is None:
            raise ThreadStopped()

        if self.creator_thread.isInterruptionRequested():
            raise ThreadStopped()
        if len(self.req_wf_chans) != 0:
            self.read_wf_files(self.selected_key)
        if self.req_data_streams != ['*']:
            not_found_data_streams = [ds for ds in self.req_data_streams
                                      if ds not in self.found_data_streams]
            if not_found_data_streams != []:
                msg = (f"No data found for data streams: "
                       f"{', '.join(map(str, not_found_data_streams))}")
                self.processing_log.append((msg, LogType.WARNING))
        self.get_gps_data()

        self.check_not_found_soh_chans()

    def read_soh_index_waveform(self, folder: str) -> None:
        """
        Loop all files in dir to read for soh data, mass position data and
            index waveform data with filename and corresponding time range
        :param folder: absolute path to data set folder
        """
        count = 0
        for path, subdirs, files in os.walk(folder):
            for file_name in files:
                if self.creator_thread.isInterruptionRequested():
                    raise ThreadStopped()
                path2file = Path(path).joinpath(file_name)
                if not validate_file(path2file, file_name):
                    continue
                if not self.read_reftek_130(path2file):
                    read_text(path2file, file_name, self.log_data['TEXT'])
                count += 1
                if count % 50 == 0:
                    self.track_info(
                        f'Read {count} file headers/ SOH files', LogType.INFO)
        self.combine_data()

    def select_key(self) -> Tuple[str, str]:
        """
        :return selected_key:
            (device's serial number, experiment_number)
            the selected keys from self.keys.
        + If there is only one key, return it.
        + If there is more than one, show all keys, let user choose one to
        return.
        """
        self.keys = sorted(list(self.keys))
        selected_key = self.keys[0]
        if len(self.keys) > 1:
            msg = ("There are more than one keys in the given data.\n"
                   "Please select one to display")
            self.pause_signal.emit(msg, self.keys)
            self.pause()
            selected_key = self.keys[self.pause_response]

        self.track_info(f'Select Key {selected_key}', LogType.INFO)
        return selected_key

    def read_wf_files(self, key: Tuple[str, str]) -> None:
        """
        :param key:
            (device's serial number, experiment_number)
            the selected keys from self.keys.
        From filesInfo, read all waveform data of requested waveform
        data_stream  channels for given sta_id, in the selected time (from Main
        Window) to append to
        DataTypeModel.__init__.waveform_data[sat_id]['read_data'][chan_id]
            ['tracesInfo'].
        tracesInfo includes list of traces' info dict which is sorted in
            startTmEpoch order by sort_data.
            refer to DataTypeModel.__init__.
            waveform_data[key]['read_data'][chan_id]['tracesInfo'].

        """
        count = 0
        for data_stream in self.waveform_data[key]['filesInfo']:
            read_data = self.waveform_data[key]['readData']
            for file_info in self.waveform_data[key]['filesInfo'][data_stream]:
                if self.creator_thread.isInterruptionRequested():
                    raise ThreadStopped()
                # file have been read
                if file_info['read']:
                    continue

                # check time
                has_data = False
                if ((self.read_start <= file_info['startEpoch'] <= self.read_end) or   # noqa: E501
                        (self.read_start <= file_info['endEpoch'] <= self.read_end)):  # noqa: E501
                    has_data = True
                if not has_data:
                    continue
                read_waveform_reftek(file_info['rt130'], key, read_data,
                                     self.data_time[key], self.tmp_dir)
                file_info['read'] = True
                count += 1
                if count % 50 == 0:
                    self.track_info(
                        f'Read {count} waveform files', LogType.INFO)
        sort_data(self.waveform_data[key]['readData'])

    def add_log(self, chan_pkt, log_info):
        if chan_pkt not in self.log_data[self.cur_key].keys():
            self.log_data[self.cur_key][chan_pkt] = []
        self.log_data[self.cur_key][chan_pkt].append(log_info)

    def read_reftek_130(self, path2file: Path) -> bool:
        """
        From the given file:
            + Read SOH data from file with SH packets,
            + read event info, mass position and index waveform (data stream)
                file from file with EH or ET packets
        :param path2file: absolute path to file
        """
        rt130 = core.Reftek130.from_file(path2file)
        unique, counts = np.unique(rt130._data["packet_type"],
                                   return_counts=True)
        nbr_packet_type = dict(zip(unique, counts))

        if b"SH" in nbr_packet_type:
            self.read_sh(path2file)
        if b"EH" in nbr_packet_type or b"ET" in nbr_packet_type:
            self.read_ehet_mp_index_wf(rt130)
        return True

    def read_sh(self, path2file: Path) -> None:
        """
        Use soh_packet library to read file with SH packet for soh data
            to append tuple (time, log string) to log_data[self.cur_key][SOH]
        :param path2file: absolute path to file
        """
        with open(path2file, "rb") as fh:
            str = fh.read()
        data = soh_packet._initial_unpack_packets_soh(str)
        for ind, d in enumerate(data):
            self.cur_key = (d['unit_id'].decode(),
                            d['experiment_number'])
            if self.cur_key not in self.log_data:
                self.log_data[self.cur_key] = {}
            logs = soh_packet.Packet.from_data(d).__str__()
            # self.add_log(d['packet_type'].decode(), (d['time'], logs))
            if 'SOH' not in self.log_data[self.cur_key]:
                self.log_data[self.cur_key]['SOH'] = []
            self.log_data[self.cur_key]['SOH'].append((d['time'], logs))

    def read_ehet_mp_index_wf(self, rt130: core.Reftek130) -> None:
        """
        Files that contents EH or ET packets are data stream files.
        There may be 1 - 9 data streams.
        :param rt130: of a data stream file in which
            + event info can be found in EH packet and save in self.log_data
            + mass position data can be found in data stream 9 and save in
                self.masspos_stream
            + waveform data can be found in data stream 1-8 which will be index
                and save to self.waveform_data[self.cur_key]["filesInfo"] to
                be processed later
        """
        data_stream = rt130._data['data_stream_number'][0] + 1
        if (self.req_data_streams != ['*'] and
                data_stream not in self.req_data_streams + [9]):
            return
        self.found_data_streams.append(data_stream)
        ind_ehet = [ind for ind, val in
                    enumerate(rt130._data["packet_type"])
                    if val in [b"EH"]]     # only need event header
        nbr_dt_samples = sum(
            [rt130._data[ind]["number_of_samples"]
             for ind in range(0, len(rt130._data))
             if ind not in ind_ehet])

        for ind in ind_ehet:
            d = rt130._data[ind]
            self.cur_key = (d['unit_id'].decode(),
                            d['experiment_number'])
            if self.cur_key not in self.log_data:
                self.log_data[self.cur_key] = {}
            logs = packet.EHPacket(d).eh_et_info(nbr_dt_samples)
            if 'EHET' not in self.log_data[self.cur_key]:
                self.log_data[self.cur_key]['EHET'] = []
            self.log_data[self.cur_key]['EHET'].append((d['time'], logs))
        if self.cur_key not in self.data_time:
            self.data_time[self.cur_key] = [constants.HIGHEST_INT, 0]
            self.keys.add(self.cur_key)
        if data_stream == 9:
            self.read_mass_pos(rt130)
        else:
            self.index_waveform(rt130, data_stream)

    def read_mass_pos(self, rt130: core.Reftek130) -> None:
        """
        Append all traces of data stream 9 to
        self.mass_pos_stream[self.currKey].
        Update data_time.
        :param rt130: for data stream 9
        """
        if self.cur_key not in self.mass_pos_stream:
            self.mass_pos_stream[self.cur_key] = Stream()

        stream = core.Reftek130.to_stream(
            rt130,
            headonly=False,
            verbose=False,
            sort_permuted_package_sequence=True)
        for tr in stream:
            self.mass_pos_stream[self.cur_key].append(tr)
            self.data_time[self.cur_key][0] = min(
                tr.stats['starttime'].timestamp,
                self.data_time[self.cur_key][0]
            )
            self.data_time[self.cur_key][1] = max(
                tr.stats['endtime'].timestamp,
                self.data_time[self.cur_key][1]
            )

    def index_waveform(self, rt130: core.Reftek130, data_stream: int) -> None:
        """
        Indexing by adding rt130 object along with time range to
            self.waveform_data[self.currKey]['filesInfo']
        :param rt130: for data stream DS
        :param data_stream: int - data stream index
        """
        if self.cur_key not in self.waveform_data:
            self.waveform_data[self.cur_key] = {"filesInfo": {},
                                                "readData": {}}
        if data_stream not in self.waveform_data[self.cur_key]["filesInfo"]:
            self.waveform_data[self.cur_key]["filesInfo"][data_stream] = []

        stream = core.Reftek130.to_stream(
            rt130,
            headonly=True,
            verbose=False,
            sort_permuted_package_sequence=True)

        self.waveform_data[self.cur_key]["filesInfo"][data_stream].append(
            {'rt130': rt130,
             'startEpoch': stream[0].stats['starttime'].timestamp,
             'endEpoch': stream[0].stats['endtime'].timestamp,
             'read': False})

    def combine_data(self) -> None:
        """
        + SOH and event log_data will be processed to create SOHdata:
            {Key of set:
                {chan_id:
                    {orgTrace: {
                        times: list of epoch time
                        data: list of channel's values
                        logIdx: list of indexes of SOH message line
                            corresponding to time and data
                    }}
                }
            }
        + massPossStream for each data set will be merged and save to
        massPossData according to chan_ids
        + Gaps will be calculate and squash for mass position only since
        currently there is no way to calculate SOH gaps, and data for waveform
        is too big to consider calculating gaps.
        """
        for k in self.log_data:
            if self.creator_thread.isInterruptionRequested():
                raise ThreadStopped()
            if k == 'TEXT':
                continue
            if k not in self.data_time:
                self.data_time[k] = [constants.HIGHEST_INT, 0]
                self.keys.add(k)
            self.soh_data[k] = {}
            logs = []
            for pkt_type in ['SOH', 'EHET']:
                if pkt_type == 'EHET':
                    logs += '\n\nEvents:'
                try:
                    for log in sorted(self.log_data[k][pkt_type],
                                      key=lambda x: x[0]):
                        # sort log data according to time
                        logs += log[1]
                        if pkt_type == 'SOH':
                            logs += '\n'
                except KeyError:
                    pass
            log_str = ''.join(logs)
            self.log_data[k] = {'SOH': [log_str]}
            log_obj = LogInfo(
                self, self.track_info, log_str, k,
                self.req_data_streams, self.req_soh_chans)
            self.data_time[k][0] = min(log_obj.min_epoch, self.data_time[k][0])
            self.data_time[k][1] = max(log_obj.max_epoch, self.data_time[k][1])
            for c_name in self.soh_data[k]:
                c = self.soh_data[k][c_name]['orgTrace']
                c['times'] = np.array(c['times'])
                c['data'] = np.array(c['data'])
                c['logIdx'] = np.array(c['logIdx'])

        all_gaps = []
        self.gaps = {k: [] for k in self.keys}
        self.mass_pos_data = {k: {} for k in self.keys}
        for k in self.mass_pos_stream:
            if self.creator_thread.isInterruptionRequested():
                raise ThreadStopped()
            stream = self.mass_pos_stream[k]
            stream.merge()
            for tr in stream:
                chan_id = tr.stats['channel']
                self.mass_pos_data[k][chan_id] = {}
                self.mass_pos_data[k][chan_id]['orgTrace'] = read_mp_trace(tr)
            gaps = stream.get_gaps()
            # gaps is list of [network, station, location, channel, starttime,
            # endtime, duration, number of missing samples]
            all_gaps += [[g[4].timestamp, g[5].timestamp] for g in gaps]
            self.gaps[k] = squash_gaps(all_gaps)

    def get_gps_data(self):
        """
        Retrieve the GPS of the current data set. Works by looking into the log
        of the data set.
        """
        log_lines = self.log_data[self.selected_key]['SOH']
        log_lines = log_lines[0].split('\n')
        log_lines = [line.strip('\r').strip('') for line in log_lines]
        gps_year = None

        for line in log_lines:
            # The lines of the log that contains the GPS data does not include
            # the year. Instead, we grab it by looking at the header of the
            # state of health packet the GPS line is in.
            # Also, it might be possible to only grab the year once. It would
            # be easy to check if the year change by seeing if the day of year
            # loops back to 1. That approach is a bit more complex, however,
            # so we are not using it.
            if 'State of Health' in line:
                current_time = line.split()[3]
                two_digit_year = current_time[:2]
                gps_year = two_digit_year_to_four_digit_year(two_digit_year)
            if 'GPS: POSITION' in line:
                self.gps_points.append(self.parse_gps_point(line, gps_year))

    @staticmethod
    def parse_gps_point(gps_line: str, gps_year: str) -> GPSPoint:
        """
        Parse a GPS log line to extract the GPS data point it contains. Needs
        to be given a year because GPS log lines do not contain the year.

        :param gps_line: the GPS log line to parse
        :param gps_year: the year associated with gps_line
        :return: a GPSPoint object that contains the GPS data points stored in
            gps_line
        """
        # RT130 data does not contain GPS fix type and number of satellites
        # used, so we set them to not available.
        fix_type = 'N/A'
        num_sats_used = 'N/A'

        time_str, *_, lat_str, long_str, height_str = gps_line.split()

        time_str = gps_year + ':' + time_str
        time_format = '%Y:%j:%H:%M:%S'
        gps_time = datetime.strptime(time_str, time_format).isoformat(sep=' ')

        # Latitude and longitude are stored in degrees, minutes, and seconds,
        # with each quantity being separated by ':'. The degree part of
        # latitude is stored as three characters, while the degree part of
        # longitude is stored as four characters. The first character of the
        # degree part is the cardinal direction (NS/EW), so we ignore it in
        # calculating latitude and longitude.
        lat_as_list = lat_str.split(':')
        lat_second = float(lat_as_list[2])
        lat_minute = float(lat_as_list[1]) + lat_second / 60
        lat = float(lat_as_list[0][1:]) + lat_minute / 60
        if lat_as_list[0][0] == 'S':
            lat = -lat

        long_as_list = long_str.split(':')
        long_second = float(long_as_list[2])
        long_minute = float(long_as_list[1]) + long_second / 60
        long = float(long_as_list[0][1:]) + long_minute / 60
        if long_as_list[0][0] == 'W':
            long = -long

        # Height is encoded as a signed float followed by the unit. We don't
        # know how many characters the unit is composed of, so we have to loop
        # through the height string backward until we can detect the end of the
        # height value.
        # Start pass the end of the string and look backward one index every
        # iteration so we don't have to add 1 to the final index.
        current_idx = len(height_str)
        current_char = height_str[current_idx - 1]
        while current_char != '.' and not current_char.isnumeric():
            current_idx -= 1
            current_char = height_str[current_idx - 1]
        height = float(height_str[:current_idx])
        height_unit = height_str[current_idx:]

        return GPSPoint(gps_time, fix_type, num_sats_used, lat, long, height,
                        height_unit)


def two_digit_year_to_four_digit_year(year: str) -> str:
    """
    Convert a year represented by two digits to its representation in 4 digits.
    Follow the POSIX and ISO C standards mentioned by the Python documentation,
    quoted below.

    'When 2-digit years are parsed, they are converted according to the POSIX
    and ISO C standards: values 69–99 are mapped to 1969–1999, and values 0–68
    are mapped to 2000–2068.'

    Raise ValueError if year is outside the 0-99 range (0n counts as n).

    :param year: the 2-digit year as a string
    :return: the 4-digit representation of year as a string
    """
    year_as_int = int(year)
    if not (0 <= year_as_int <= 99):
        raise ValueError(f'Given year {year} is not in the valid range.')
    if int(year) < 69:
        prefix = '20'
    else:
        prefix = '19'
    return f'{prefix}{year:0>2}'
